/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action delete_note(doc *d, int line_number, int chord_number, int note)
{
  data {
    int line_number;
    int chord_number;
    int note;
    elem e;
  }

  do {
    data->line_number = line_number;
    data->chord_number = chord_number;
    data->note = note;
    data->e = d->data[line_number].u.music.data[chord_number].
              u.chord.data[note];

    if (!doc_delete_note(d, line_number, chord_number, note)) return 0;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      /* put the note back in place */
      switch (data->e.type) {
        case NOTE:
          doc_insert_note(d, line_number, chord_number, data->e.type,
                          data->e.u.note.pitch,
                          data->e.u.note.alteration,
                          data->e.u.note.duration,
                          data->e.u.note.nb_dots,
                          data->e.u.note.head_delta,
                          data->e.u.note.alteration_delta,
                          data->e.u.note.dots_delta,
                          data->e.u.note.stem_down);
          break;
        case SILENCE:
          doc_insert_note(d, line_number, chord_number, data->e.type,
                          data->e.u.silence.position,
                          data->e.u.silence.duration,
                          data->e.u.silence.nb_dots);
          break;
      }
      return 0;
    }

    d->state = STATE_MODIFIED;
    size_change_chord(d, line_number, chord_number);

    return 1;
  }

  undo {
    /* put the note back in place */
    switch (data->e.type) {
      case NOTE:
        doc_insert_note(d, data->line_number, data->chord_number,
                        data->e.type,
                        data->e.u.note.pitch,
                        data->e.u.note.alteration,
                        data->e.u.note.duration,
                        data->e.u.note.nb_dots,
                        data->e.u.note.head_delta,
                        data->e.u.note.alteration_delta,
                        data->e.u.note.dots_delta,
                        data->e.u.note.stem_down);
        break;
      case SILENCE:
        doc_insert_note(d, data->line_number, data->chord_number,
                        data->e.type,
                        data->e.u.silence.position,
                        data->e.u.silence.duration,
                        data->e.u.silence.nb_dots);
        break;
    }
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_change_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  redo {
    doc_delete_note(d, data->line_number, data->chord_number, data->note);
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_change_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
  }
}
