/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action insert_chord(doc *d, int line_number, int before_chord,
                    chord_type type)
{
  data {
    int line_number;
    int chord_number;
    chord_type ch;
    chord c;
  }

  do {
    data->line_number = line_number;
    if (before_chord == -1)
      data->chord_number = d->data[line_number].u.music.size;
    else
      data->chord_number = before_chord;
    data->ch = type;

    /* for now on the chord is "alive", we store nothing */
    data->c.u.chord.data = NULL;
    data->c.u.chord.size = 0;
    data->c.u.chord.maxsize = 0;

    if (!doc_insert_chord(d, line_number, before_chord, type)) return 0;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      doc_delete_chord(d, line_number, data->chord_number);
      return 0;
    }

    size_new_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;

    return 1;
  }

  undo {
    /* "save" the chord and let the app. "forget" it */
    chord *c = &d->data[data->line_number].u.music.data[data->chord_number];
    data->c = *c;
    if (c->type == NOTES_CHORD) {
      c->u.chord.data = NULL;
      c->u.chord.size = 0;
      c->u.chord.maxsize = 0;
    }
    doc_delete_chord(d, data->line_number, data->chord_number);
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_delete_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  redo {
    if (data->chord_number == d->data[data->line_number].u.music.size)
      doc_insert_chord(d, data->line_number, -1, data->ch);
    else
      doc_insert_chord(d, data->line_number, data->chord_number, data->ch);
    /* let the app. "remember" the chord */
    d->data[data->line_number].u.music.data[data->chord_number] = data->c;
    /* we have to forget it */
    data->c.u.chord.data = NULL;
    data->c.u.chord.size = 0;
    data->c.u.chord.maxsize = 0;
    d->cursor.line = data->line_number;
    d->cursor.chord = data->chord_number;
    size_new_chord(d, data->line_number, data->chord_number);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    destroy_chord(&data->c);
  }
}
