/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action insert_music_line(doc *d, int after_line)
{
  data {
    int line;
    void *data;
    int maxsize;
  }

  do {
    data->line = after_line+1;

    size_prenew_line(d, after_line+1);

    if (!doc_insert_line(d, LINE_MUSIC, after_line))
      return 0;

    data->data = NULL;
    data->maxsize = 0;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      doc_delete_line(d, data->line);
      return 0;
    }

    size_new_line(d, data->line);
    d->state = STATE_MODIFIED;

    return 1;
  }

  undo {
    /* "remember" the soon deleted line */
    data->data = d->data[data->line].u.music.data;
    data->maxsize = d->data[data->line].u.music.maxsize;
    size_predelete_line(d, data->line);
    d->data[data->line].u.music.data = NULL;
    d->data[data->line].u.music.maxsize = 0;
    doc_delete_line(d, data->line);
    d->cursor.line = data->line >= d->size ? d->size-1 : data->line;
    d->cursor.chord = 0;
    size_delete_line(d, data->line);
    d->state = STATE_MODIFIED;
  }

  redo {
    size_prenew_line(d, data->line-1);
    doc_insert_line(d, LINE_MUSIC, data->line-1);
    /* set the data back to the app */
    d->data[data->line].u.music.data = data->data;
    d->data[data->line].u.music.maxsize = data->maxsize;
    /* and forget it */
    data->data = NULL;
    data->maxsize = 0;
    d->cursor.line = data->line;
    d->cursor.chord = 0;
    size_new_line(d, data->line);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    if (data->maxsize) {
      free(data->data);
      data->data = NULL;
      data->maxsize = 0;
    }
  }
}
