/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action insert_text_line(doc *d, int after_line)
{
  data {
    int line;
    void *text;
  }

  do {
    char *t = strdup("Text line");
    data->line = after_line+1;
    data->text = NULL;

    if (!t) {
      set_error(OOM);
      return 0;
    }

    size_prenew_line(d, after_line);

    if (!doc_insert_line(d, LINE_TEXT, after_line)) {
      free(t);
      return 0;
    }
    LINE_TEXT_DATA(d, data->line) = t;

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      free(t);
      LINE_TEXT_DATA(d, data->line) = NULL;
      doc_delete_line(d, data->line);
      return 0;
    }

    size_new_line(d, data->line);
    d->state = STATE_MODIFIED;

    return 1;
  }

  undo {
    /* remember text */
    data->text = LINE_TEXT_DATA(d, data->line);
    LINE_TEXT_DATA(d, data->line) = NULL;
    size_predelete_line(d, data->line);
    doc_delete_line(d, data->line);
    d->cursor.line = data->line >= d->size ? d->size-1 : data->line;
    d->cursor.chord = 0;
    size_delete_line(d, data->line);
    d->state = STATE_MODIFIED;
  }

  redo {
    size_prenew_line(d, data->line-1);
    doc_insert_line(d, LINE_TEXT, data->line-1);
    /* restore text */
    LINE_TEXT_DATA(d, data->line) = data->text;
    data->text = NULL;
    d->cursor.line = data->line;
    d->cursor.chord = 0;
    size_new_line(d, data->line);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    if (data->text) {
      free(data->text);
      data->text = NULL;
    }
  }
}
