/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

action merge_lines(doc *d, int first_line)
{
  data {
    int first_line;
    int chord_pos;
    void *data;
    int maxsize;
  }

  do {
    line *l = &d->data[first_line+1];
    data->first_line = first_line;
    data->chord_pos = d->data[first_line].u.music.size;
    data->data = l->u.music.data;
    data->maxsize = l->u.music.maxsize;

    size_premerge_line(d, first_line+1);

    /* this function will automatically "forget" the line in the data */
    if (!doc_merge_lines(d, first_line, MERGE_DONT_DELETE)) {
      /* size_premerge_line modifies the global height, we need to reset */
      size_reset(d);
      return 0;
    }

    if (REGISTER_HISTORY < 0) {
      set_error(OOM);
      doc_split_line(d, data->first_line, data->chord_pos,
                     data->data, data->maxsize);
      /* size_premerge_line modifies the global height, we need to reset */
      size_reset(d);
      return 0;
    }

    size_merge_line(d, data->first_line);
    d->state = STATE_MODIFIED;

    return 1;
  }

  undo {
    size_presplit_line(d, data->first_line);
    doc_split_line(d, data->first_line, data->chord_pos,
                   data->data, data->maxsize);
    /* "forget" second line, it's used again by the application */
    data->data = NULL;
    data->maxsize = 0;
    d->cursor.line = data->first_line + 1;
    d->cursor.chord = 0;
    size_split_line(d, data->first_line);
    d->state = STATE_MODIFIED;
  }

  redo {
    /* "remember" again the 2nd line, the app. will "forget" it */
    data->data = d->data[data->first_line+1].u.music.data;
    data->maxsize = d->data[data->first_line+1].u.music.maxsize;
    size_premerge_line(d, data->first_line+1);
    doc_merge_lines(d, data->first_line, MERGE_DONT_DELETE);
    d->cursor.line = data->first_line;
    d->cursor.chord = data->chord_pos;
    size_merge_line(d, data->first_line);
    d->state = STATE_MODIFIED;
  }

  save {
  }

  load {
  }

  erase {
    if (data->maxsize) {
      free(data->data);
      data->data = NULL;
      data->maxsize = 0;
    }
  }
}
