/* a barbarian elf symbols' table extractor
 * public domain stuff
 * part of xgprof
 * (available at http://sed.free.fr)
 * if bugs are there, contact sed@free.fr for them to be removed
 * not all targets are supported, if you need support for your target
 * contact the author as well
 *
 * the ABI's description was found at:
 * http://www.sco.com/developers/gabi/latest
 * (yeah, shame on me for using SCO stuff, but the heck...)
 */

#include "elf.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>

#ifdef USE_MMAP
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#endif /* USE_MMAP */

#include "sstring.h"
#include "symlist.h"

int big_endian;
int bit_mode = 32;

static int read_char(FILE *f, unsigned long *ret)
{
  int c;
  c = fgetc(f);
  if (c == EOF) return -1;
  *ret = c&255;
  return 0;
}

static int read_half(FILE *f, unsigned long *ret)
{
  int c;
  c = fgetc(f);
  if (c == EOF) return -1;
  if (big_endian)
    *ret = (c&255) << 8;
  else
    *ret = c&255;
  c = fgetc(f);
  if (c == EOF) return -1;
  if (big_endian)
    *ret |= c&255;
  else
    *ret |= (c&255)<<8;
  return 0;
}

static int read_word(FILE *f, unsigned long *ret)
{
  int c;
  c = fgetc(f);
  if (c == EOF) return -1;
  if (big_endian) {
    *ret = (c&255) << 24;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= (c&255) << 16;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= (c&255) << 8;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= c&255;
  } else {
    *ret = c&255;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= (c&255) << 8;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= (c&255) << 16;
    c = fgetc(f);
    if (c == EOF) return -1;
    *ret |= (c&255) << 24;
  }
  return 0;
}

static int read_xword(FILE *f, uint64_t *ret)
{
  int c;
  uint64_t dum;
  c = fgetc(f);
  if (c == EOF) return -1;
  if (big_endian) {
    dum = c&255; dum <<= 32+24;
    *ret = dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32+16;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32+8;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 24;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 16;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 8;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255;
    *ret |= dum;
  } else {
    dum = c&255;
    *ret = dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 8;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 16;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 24;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32+8;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32+16;
    *ret |= dum;
    c = fgetc(f);
    if (c == EOF) return -1;
    dum = c&255; dum <<= 32+24;
    *ret |= dum;
  }
  return 0;
}

static int read_addr32(FILE *f, unsigned long *ret)
{
  return read_word(f, ret);
}

static int read_addr64(FILE *f, uint64_t *ret)
{
  return read_xword(f, ret);
}

static int read_offset32(FILE *f, unsigned long *ret)
{
  return read_word(f, ret);
}

static int read_offset64(FILE *f, uint64_t *ret)
{
  return read_xword(f, ret);
}

#define READ(type, var) \
  do { \
    if (read_ ## type(f, &var) == -1) { \
      fprintf(stderr, "%s: %s\n", n, strerror(errno)); \
      goto elf_err; \
    } \
  } while (0)

#define READ_SECTION(var) \
  do { \
    if (bit_mode == 32) { \
      READ(word, var.sh_name); \
      READ(word, var.sh_type); \
      READ(word, var.sh_flags32); \
      READ(addr32, var.sh_addr32); \
      READ(offset32, var.sh_offset32); \
      READ(word, var.sh_size32); \
      READ(word, var.sh_link); \
      READ(word, var.sh_info); \
      READ(word, var.sh_addralign32); \
      READ(word, var.sh_entsize32); \
    } else { \
      READ(word, var.sh_name); \
      READ(word, var.sh_type); \
      READ(xword, var.sh_flags64); \
      READ(addr64, var.sh_addr64); \
      READ(offset64, var.sh_offset64); \
      READ(xword, var.sh_size64); \
      READ(word, var.sh_link); \
      READ(word, var.sh_info); \
      READ(xword, var.sh_addralign64); \
      READ(xword, var.sh_entsize64); \
    } \
  } while (0)

#define READ_SYMBOL32(var) \
  do { \
    READ(word, var.name); \
    READ(addr32, var.value32); \
    READ(word, var.size32); \
    READ(char, var.info); \
    READ(char, var.other); \
    READ(half, var.shndx); \
  } while (0)

#define READ_SYMBOL64(var) \
  do { \
    READ(word, var.name); \
    READ(char, var.info); \
    READ(char, var.other); \
    READ(half, var.shndx); \
    READ(addr64, var.value64); \
    READ(xword, var.size64); \
  } while (0)

#define READ_SYMBOL(var) \
  do { \
    if (bit_mode == 32) { \
      READ_SYMBOL32(var); \
    } else { \
      READ_SYMBOL64(var); \
    } \
  } while (0)

typedef struct {
    unsigned long sh_name;
    unsigned long sh_type;
    unsigned long sh_flags32;
    uint64_t sh_flags64;
    unsigned long sh_addr32;
    uint64_t sh_addr64;
    unsigned long sh_offset32;
    uint64_t sh_offset64;
    unsigned long sh_size32;
    uint64_t sh_size64;
    unsigned long sh_link;
    unsigned long sh_info;
    unsigned long sh_addralign32;
    uint64_t sh_addralign64;
    unsigned long sh_entsize32;
    uint64_t sh_entsize64;
} section;

typedef struct {
  unsigned long name;
  unsigned long value32;
  uint64_t value64;
  unsigned long size32;
  uint64_t size64;
  unsigned long info;
  unsigned long other;
  unsigned long shndx;
} symbol;

/* We can do globals since we read only one file (the heck if you
 * want to use this code in your project, go hack the gnu stuff
 * instead) (they also use globals, HAHAHA!)
 */
static char *function_name;
static int function_size;
static int function_maxsize;

static int put_char(int c)
{
  if (function_size == function_maxsize) {
    char *new;
    int ns = function_maxsize + 1024;
    if (function_maxsize)
      new = realloc(function_name, ns);
    else
      new = malloc(ns);
    if (!new) {
      fprintf(stderr, "out of memory\n");
      return -1;
    }
    function_name = new;
    function_maxsize = ns;
  }

  function_name[function_size] = c;
  function_size++;

  return 0;
}

#ifdef USE_MMAP
static char *names;
static size_t names_size;
#endif

static int get_function_name(char *n, FILE *f, unsigned long offset)
{
#ifndef USE_MMAP
  int c;
  function_size = 0;
  if (fseek(f, offset, SEEK_SET) == -1) {
    perror(n);
    return -1;
  }
  while (1) {
    c = fgetc(f);
    if (c == 0 || c == EOF) break;
    if (put_char(c) == -1) return -1;
  }
#else /* USE_MMAP */
  function_size = 0;
  while (1) {
    if (names[offset] == 0) break;
    if (put_char(names[offset]) == -1) return -1;
    offset++;
  }
#endif /* USE_MMAP */
  return put_char(0);
}

static int elf_add_symbol(char *file_name, FILE *f,
                          symbol *s, unsigned long names_offset)
{
  if (bit_mode == 32) if (s->value32 == 0 || s->size32 == 0) return 0;
  if (bit_mode == 64) if (s->value64 == 0 || s->size64 == 0) return 0;

  if (get_function_name(file_name, f, names_offset+s->name) == -1)
    return -1;

  if (bit_mode == 32) return add_symbol(s->value32, s->size32, function_name);

  /* for now we require higher 32b to be 0 if sizeof(unsigned long) < 8
   * (since 64b support came late, the outside world keeps using unsigned
   * long which is only 32b on some hosts)
   * (yeah, bad hacking, but anyway who will analyze a 64b binary on a
   * 32b host, huh?)
   */
  if (sizeof(unsigned long) < 8) {
    if ((s->value64 & (uint64_t)(0xfffffffflu)) != s->value64 ||
        (s->size64 & (uint64_t)(0xfffffffflu)) != s->size64) {
      fprintf(stderr,
              "ERROR: elf file not supported because of 32b/64b stuff\n");
      fprintf(stderr, "ERROR: contact sed@free.fr to fix this problem.\n");
      return -1;
    }
  }

  return add_symbol(s->value64, s->size64, function_name);
}

int get_elf_symbols(char *n)
{
  FILE *f = fopen(n, "r");
  unsigned char head[16];
  unsigned long type;
  unsigned long machine;
  unsigned long version;
  unsigned long entry32;
  unsigned long phoff32;
  unsigned long shoff32;
  uint64_t entry64;
  uint64_t phoff64;
  uint64_t shoff64;
  unsigned long flags;
  unsigned long ehsize;
  unsigned long phentsize;
  unsigned long phnum;
  unsigned long shentsize;
  unsigned long shnum;
  unsigned long shstrndx;
  section symtab;
  section strtab;
  int i;
  unsigned long nb_symbols;
#ifdef USE_MMAP
  unsigned long names_offset;
#endif

  if (!f) goto elf_perr;
  if (fread(head, 1, 16, f) != 16) goto elf_perr;
  if (head[0] != 0x7f || head[1] != 'E' || head[2] != 'L' || head[3] != 'F') {
    fprintf(stderr, "%s: no ELF header found\n", n);
    goto elf_err;
  }
  /* class: 0: invalid, 1: 32b, 2: 64b */
  if (head[4] != 1 && head[4] != 2) {
    fprintf(stderr, "%s: ELF class (%d) not supported\n", n, head[4]);
    goto elf_err;
  }
  if (head[4] == 2) bit_mode = 64; else bit_mode = 32;
  /* endianness */
  big_endian = head[5];
  if (big_endian != 1 && big_endian != 2) {
    fprintf(stderr, "%s: unsupported endianness (%d)\n", n, head[5]);
    goto elf_err;
  }
  big_endian--;
  /* version */
  if (head[6] != 1) {
    fprintf(stderr, "%s: ELF version (%d) not supported\n", n, head[6]);
    goto elf_err;
  }
  /* os ABI, let's ignore it */
  /* ABI version, let's also ignore it */

  /* get the header */
  READ(half, type);
  READ(half, machine);
  READ(word, version);
  if (bit_mode == 32) {
    READ(addr32, entry32);
    READ(offset32, phoff32);
    READ(offset32, shoff32);
  } else {
    READ(addr64, entry64);
    READ(offset64, phoff64);
    READ(offset64, shoff64);
  }
  READ(word, flags);
  READ(half, ehsize);
  READ(half, phentsize);
  READ(half, phnum);
  READ(half, shentsize);
  READ(half, shnum);
  READ(half, shstrndx);
  if (version != 1) {
    fprintf(stderr, "%s: bad ELF version (%lu)\n", n, version);
    goto elf_err;
  }
  if (type != 2) {
    fprintf(stderr, "%s: not an executable (type is %lu)\n", n, type);
    goto elf_err;
  }
  if ((bit_mode == 32 && shoff32 == 0) ||
      (bit_mode == 64 && shoff64 == 0)) {
    fprintf(stderr, "%s: no sections in ELF file\n", n);
    goto elf_err;
  }
  if (shnum == 0) {
    fprintf(stderr,
        "%s: shnum=0, heck, contact sed@free.fr, this case must be handled\n",
        n);
    goto elf_err;
  }

  /* get the symbol table */
  symtab.sh_type = 0;
  for (i = 0; i < shnum; i++) {
    section s;

    if (fseek(f, bit_mode == 32 ?
                   shoff32 + i * shentsize :
                   shoff64 + i * shentsize,
              SEEK_SET) == -1) {
      perror(n);
      goto elf_err;
    }

    READ_SECTION(s);

    if (s.sh_type == 2) {
      if (symtab.sh_type == 2) {
        fprintf(stderr, "%s: more than one symbol table\n", n);
        goto elf_err;
      }
      symtab = s;
    }
  }
  if (symtab.sh_type == 0) {
    fprintf(stderr,
        "%s: no symbol table in ELF file (did you strip the binary?)\n", n);
    goto elf_err;
  }
  if ((bit_mode == 32 && symtab.sh_size32 == 0) ||
      (bit_mode == 64 && symtab.sh_size64 == 0)) {
    fprintf(stderr, "%s: symbol table of size null\n", n);
    goto elf_err;
  }
  if (bit_mode == 32 &&
       (symtab.sh_entsize32 == 0 ||
        symtab.sh_size32 % symtab.sh_entsize32 != 0)) {
    fprintf(stderr,
        "%s: invalid symbol table entry size / total size (%lu / %lu)\n", n,
        symtab.sh_entsize32, symtab.sh_size32);
    goto elf_err;
  }
  if (bit_mode == 64 &&
       (symtab.sh_entsize64 == 0 ||
        symtab.sh_size64 % symtab.sh_entsize64 != 0)) {
    fprintf(stderr,
        "%s: invalid symbol table entry size / total size (%lu / %lu)\n", n,
        (unsigned long)symtab.sh_entsize64, (unsigned long)symtab.sh_size64);
    goto elf_err;
  }
  if (bit_mode == 32) {
    nb_symbols = symtab.sh_size32 / symtab.sh_entsize32;
  } else {
    nb_symbols = symtab.sh_size64 / symtab.sh_entsize64;
  }

  /* get the string table of the symbol table */
  if (fseek(f, bit_mode == 32 ?
                 shoff32 + shentsize * symtab.sh_link :
                 shoff64 + shentsize * symtab.sh_link,
            SEEK_SET) == -1) {
    perror(n);
    goto elf_err;
  }
  READ_SECTION(strtab);
  if (strtab.sh_type != 3) {
    fprintf(stderr, "%s: wrong string table for symbol table\n", n);
    goto elf_err;
  }

  /* let's read symbols */
#ifdef USE_MMAP
#ifndef MAP_FAILED
#define MAP_FAILED (void*)(-1L)
#endif /* MAP_FAILED */
{
  /* we must align the mmap to PAGESIZE boundary */
  unsigned long pagesize;
  off_t offset;
  size_t len;
  off_t real_offset;
  size_t real_len;
  struct stat buf;

  if (fstat(fileno(f), &buf) == -1) {
    perror("fstat");
    goto elf_err;
  }

  pagesize = buf.st_blksize;

  offset = bit_mode == 32 ? strtab.sh_offset32 : strtab.sh_offset64;
  len = bit_mode == 32 ? strtab.sh_size32 : strtab.sh_size64;
  real_offset = offset & (~(pagesize-1));
  real_len = len + (offset-real_offset);

  if ((names=mmap(NULL, real_len,
           PROT_READ, MAP_SHARED, fileno(f),
           real_offset)) ==
        MAP_FAILED) {
    perror("mmap");
    goto elf_err;
  }
  names_size = real_len;
  names_offset = offset-real_offset;
}

  if (fseek(f, bit_mode == 32 ?
                 symtab.sh_offset32 :
                 symtab.sh_offset64,
            SEEK_SET) == -1) {
    perror(n);
    goto elf_err;
  }
#endif /* USE_MMAP */

  for (i = 0; i < nb_symbols; i++) {
    symbol s;
    int type;

#ifndef USE_MMAP
    if (fseek(f, bit_mode == 32 ?
                   symtab.sh_offset32 + i * symtab.sh_entsize32 :
                   symtab.sh_offset64 + i * symtab.sh_entsize64,
              SEEK_SET) == -1) {
      perror(n);
      goto elf_err;
    }
#endif /* USE_MMAP */

    READ_SYMBOL(s);

    type = s.info & 0xf;

    /* get the functions */
    if (type == 2)
      if (elf_add_symbol(n, f, &s,
#ifdef USE_MMAP
                         names_offset
#else
                         bit_mode == 32 ?
                           strtab.sh_offset32 :
                           strtab.sh_offset64
#endif
                        ) == -1) goto elf_err;
  }

#ifdef USE_MMAP
  munmap(names, names_size);
#endif /* USE_MMAP */

  fclose(f);
  return 0;

elf_perr:
    perror(n);
elf_err:
    if(f) fclose(f);
    return -1;
}

void free_elf(void)
{
  if (function_maxsize) {
    free(function_name);
    function_name = NULL;
    function_size = function_maxsize = 0;
  }
}
