/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

/* doc for this stuff is in the file GREY */

#include "grey.h"

#include <stdlib.h>

#include "gtk_wrapper.h"
#include "dprintf.h"

static grey_item data[__GREY_NB_ITEMS];

void greyreg(grey_type t, void *p)
{
  /* the dprintf cannot use the name of the function because of 'make grey'
   * that would pick up the line, thus the %c
   */
  dprintf("greyre%c %d %p\n", 'g', t, p);
  data[t].level = 1;
  data[t].pointer = p;
  gtk_wrapper_set_sensitive(p, 0);
}

void greyunreg(grey_type t)
{
  dprintf("greyunreg %d\n", t);
  data[t].pointer = NULL;
}

void grey(grey_type t)
{
  dprintf("grey %d (level=%d pointer=%p)\n", t,
           data[t].level, data[t].pointer);
  data[t].level++;
  if (data[t].level == 1) gtk_wrapper_set_sensitive(data[t].pointer, 0);
}

void ungrey(grey_type t)
{
  dprintf("ungrey %d (level=%d pointer=%p)\n", t,
           data[t].level, data[t].pointer);
  if (data[t].level == 1) gtk_wrapper_set_sensitive(data[t].pointer, 1);
  data[t].level--;
}

void greyall(void)
{
  int i;
  dprintf("greyall\n");
  for (i = 0; i < __GREY_NB_ITEMS; i++)
    if (data[i].pointer) grey(i);
}

void ungreyall(void)
{
  int i;
  dprintf("ungreyall\n");
  for (i = 0; i < __GREY_NB_ITEMS; i++)
    if (data[i].pointer) ungrey(i);
}

void set_grey_undo_redo(doc *d)
{
  int old_undo = data[GREY_MENU_UNDO].level;
  int old_redo = data[GREY_MENU_REDO].level;
  dprintf("set_grey_undo_redo\n");
  data[GREY_MENU_UNDO].level =
    (history_get_pos(&d->history) == 0 ||
     history_get_pos(&d->history) == -1) ? 1 : 0;
  data[GREY_MENU_REDO].level =
    history_get_len(&d->history)-1 == history_get_pos(&d->history) ? 1 : 0;
  if (old_undo != data[GREY_MENU_UNDO].level)
    gtk_wrapper_set_sensitive(data[GREY_MENU_UNDO].pointer,
      !data[GREY_MENU_UNDO].level);
  if (old_redo != data[GREY_MENU_REDO].level)
    gtk_wrapper_set_sensitive(data[GREY_MENU_REDO].pointer,
      !data[GREY_MENU_REDO].level);
}

void forcegrey(grey_type t, int level)
{
  dprintf("forcegrey %d %d (pointer %p)\n", t, level, data[t].pointer);
  int old_level = data[t].level;
  data[t].level = level;
  /* no change? */
  if ((level > 0 && old_level > 0) ||
      (level <= 0 && old_level <= 0)) return;
  gtk_wrapper_set_sensitive(data[t].pointer, level > 0 ? 0 : 1);
}
