/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "gtk_wrapper.h"

#include <gtk/gtk.h>
#include <stdlib.h>

#include "warning_gui.h"
#include "dprintf.h"

static GdkColor colors[NB_COLORS];
static GdkGC *gc;
static int current_color = COLOR_BLACK;

/* setup in images.c */
GdkPixmap *stipples[4][NB_IMAGES];
int stipples_width[4][NB_IMAGES];
int stipples_height[4][NB_IMAGES];

void gtk_wrapper_window_set_modal(void *win, int bool)
{
  gtk_window_set_modal(GTK_WINDOW(win), bool ? TRUE : FALSE);
}

void gtk_wrapper_set_sensitive(void *p, int s)
{
  gtk_widget_set_sensitive(GTK_WIDGET(p), s ? TRUE : FALSE);
}

void gtk_wrapper_widget_show_all(void *w)
{
  gtk_widget_show_all(w);
}

void gtk_wrapper_main(void)
{
  gtk_main();
}

void gtk_wrapper_create_gc(void *win)
{
  gc = gdk_gc_new(((GtkWidget *)win)->window);
  gdk_gc_copy(gc, ((GtkWidget*)win)->style->fg_gc[GTK_STATE_NORMAL]);
}

#ifdef DEBUG_ON
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
int toto(void *u, GdkEventKey *ev)
{
  if (ev->type == GDK_KEY_RELEASE) return 0;
  if (ev->keyval == GDK_Return)
    printf("KEY return\n");
  else if (ev->keyval == GDK_BackSpace)
    printf("KEY backspace\n");
  else if (ev->keyval == GDK_Left)
    printf("KEY left\n");
  else if (ev->keyval == GDK_Right)
    printf("KEY right\n");
  else if (ev->keyval == GDK_Up)
    printf("KEY up\n");
  else if (ev->keyval == GDK_Down)
    printf("KEY down\n");
  else if (ev->keyval == 122)
    printf("KEY ctrl+z '%s'\n", ev->string);
  else if (ev->keyval == 121)
    printf("KEY ctrl+y '%s'\n", ev->string);
  else if (ev->length)
    printf("KEY '%s' (%u)\n", ev->string, ev->keyval);
  fflush(stdout);
  return 0;
}
#endif

void gtk_wrapper_init(int *n, char ***v)
{
  PangoColor col;

  gtk_init(n, v);

#ifdef DEBUG_ON
  /* some debugging facility */
  gtk_key_snooper_install((GtkKeySnoopFunc)toto, 0);
#endif

  /* create the colors */
  /* green */
  if (!pango_color_parse(&col, "green")) goto color_error;
  colors[COLOR_GREEN].red = col.red;
  colors[COLOR_GREEN].green = col.green;
  colors[COLOR_GREEN].blue = col.blue;
  if (!gdk_colormap_alloc_color(gdk_colormap_get_system(),
                                &colors[COLOR_GREEN], FALSE, TRUE))
    goto color_error;
  /* blue */
  if (!pango_color_parse(&col, "blue")) goto color_error;
  colors[COLOR_BLUE].red = col.red;
  colors[COLOR_BLUE].green = col.green;
  colors[COLOR_BLUE].blue = col.blue;
  if (!gdk_colormap_alloc_color(gdk_colormap_get_system(),
                                &colors[COLOR_BLUE], FALSE, TRUE))
    goto color_error;
  /* black */
  if (!pango_color_parse(&col, "black")) goto color_error;
  colors[COLOR_BLACK].red = col.red;
  colors[COLOR_BLACK].green = col.green;
  colors[COLOR_BLACK].blue = col.blue;
  if (!gdk_colormap_alloc_color(gdk_colormap_get_system(),
                                &colors[COLOR_BLACK], FALSE, TRUE))
    goto color_error;
  /* white */
  if (!pango_color_parse(&col, "white")) goto color_error;
  colors[COLOR_WHITE].red = col.red;
  colors[COLOR_WHITE].green = col.green;
  colors[COLOR_WHITE].blue = col.blue;
  if (!gdk_colormap_alloc_color(gdk_colormap_get_system(),
                                &colors[COLOR_WHITE], FALSE, TRUE))
    goto color_error;

#if 0
{
  GtkClipboard *b = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
  if (!b) {
    printf("error with gtk_clipboard_get\n");
    exit(1);
  }
  GtkTargetEntry targets;
  targets.target = "TEXT";
  targets.flags = 0;
  targets.info = 0xfecf;
  void get_fun(GtkClipboard *b, GtkSelectionData *d, unsigned info, void *p)
  {
    printf("get fun\n");
    gtk_selection_data_set_text(d, "test", 4);
  }
  void clear_fun(GtkClipboard *b, void *d)
  {
    printf("clear fun\n");
  }
  gtk_clipboard_set_with_data(b, &targets, 1, get_fun, clear_fun, (void *)0);
  gtk_clipboard_set_text(b, "fuck", 4);
  gtk_main();
  exit(0);
}
#endif

  return;

color_error:
  warning(
    "Unable to allocate colors, behavior of the program will be undefined\n",
    0);
}

void gtk_wrapper_main_quit(void)
{
  gtk_main_quit();
}

void gtk_wrapper_widget_destroy(void *w)
{
  gtk_widget_destroy(w);
}

void gtk_wrapper_widget_queue_draw(void *w)
{
  dprintf("gtk_wrapper_widget_queue_draw\n");
  gtk_widget_queue_draw(w);
}

void gtk_wrapper_widget_queue_draw_area(void *s, int x, int y, int w, int h,
                                        int zoom_level)
{
  dprintf("gtk_wrapper_widget_queue_draw_area x=%d y=%d w=%d h=%d\n",x,y,w,h);
  gtk_widget_queue_draw_area(s, x * zoom_level, y * zoom_level,
                             w * zoom_level, h * zoom_level);
}

void gtk_wrapper_label_set_text(void *w, const char *t)
{
  gtk_label_set_text(GTK_LABEL(w), t);
}

void gtk_wrapper_entry_set_text(void *w, const char *t)
{
  gtk_entry_set_text(GTK_ENTRY(w), t);
}

void gtk_wrapper_draw_line(void *d, int x1, int y1, int x2, int y2,
                           int zoom_level)
{
  int i;
  /* only horizontal and vertical lines allowed */
  if (x1 == x2) {
    if (y1 > y2) {
      int dum;
      dum = y1; y1 = y2; y2 = dum;
      dum = x1; x1 = x2; x2 = dum;
    }
    for (i = 0; i < zoom_level; i++)
      gdk_draw_line(((GtkWidget *)d)->window, gc,
                     x1 * zoom_level + i, y1 * zoom_level,
                     x2 * zoom_level + i, y2 * zoom_level + zoom_level - 1);
  }

  if (x1 > x2) {
    int dum;
    dum = y1; y1 = y2; y2 = dum;
    dum = x1; x1 = x2; x2 = dum;
  }
  for (i = 0; i < zoom_level; i++)
    gdk_draw_line(((GtkWidget *)d)->window, gc,
                   x1 * zoom_level, y1 * zoom_level + i,
                   x2 * zoom_level + zoom_level - 1, y2 * zoom_level + i);
}

/* gcomposer should cache the images on the X server, this is too slow */
void gtk_wrapper_draw_points(void *d, void *pts, int nb_pts, int dx, int dy,
                             int zoom_level)
{
  /* we need to rewrite this code, too much calculations */
  GdkPoint *spts = pts;
  GdkPoint tpts[nb_pts * zoom_level * zoom_level];
  int i;
  int pt;
  int x, y;
  for (i = 0, pt = 0; i < nb_pts; i++) {
    for (x = 0; x < zoom_level; x++) for (y = 0; y < zoom_level; y++, pt++) {
      tpts[pt].x = spts[i].x * zoom_level + dx * zoom_level + x;
      tpts[pt].y = spts[i].y * zoom_level + dy * zoom_level + y;
    }
  }
  gdk_draw_points(((GtkWidget *)d)->window,
                  /*((GtkWidget *)d)->style->fg_gc[GTK_STATE_NORMAL]*/ gc,
                  tpts, nb_pts * zoom_level * zoom_level);
}

void gtk_wrapper_draw_image(void *d, enum image_type t, int x, int y,
                            int zoom)
{
#if 0
  GdkPixmap *p;
  switch (current_color) {
    case COLOR_BLACK: p = pixmaps_black[zoom-1][t]; break;
    case COLOR_BLUE: p = pixmaps_blue[zoom-1][t]; break;
    case COLOR_GREEN: p = pixmaps_green[zoom-1][t]; break;
  }

  gdk_draw_pixmap(((GtkWidget*)d)->window, gc, p, 0, 0,
                  x * zoom, y * zoom, -1, -1);
#endif

  x *= zoom;
  y *= zoom;

  gdk_gc_set_stipple(gc, stipples[zoom-1][t]);
  gdk_gc_set_fill(gc, GDK_STIPPLED);
  gdk_gc_set_ts_origin(gc, x, y);
  gdk_draw_rectangle(((GtkWidget *)d)->window, gc, TRUE, x, y,
                     stipples_width[zoom-1][t],
                     stipples_height[zoom-1][t]);
  gdk_gc_set_fill(gc, GDK_SOLID);
}

void *gtk_wrapper_alloc_points(int n)
{
  return malloc(sizeof(GdkPoint) * n);
}

void gtk_wrapper_add_point(void *pts, int cur_point, int x, int y)
{
  GdkPoint *t = pts;
  t[cur_point].x = x;
  t[cur_point].y = y;
}

void gtk_wrapper_source_remove(unsigned int t)
{
  g_source_remove(t);
}

unsigned int gtk_wrapper_timeout_add(unsigned int t, void *a, void *b)
{
  return g_timeout_add(t, (GSourceFunc)a, (gpointer)b);
}

void gtk_wrapper_drawing_area_set_dimension(void *d, int w, int h,
                                            int zoom_level)
{
  dprintf("gtk_wrapper_drawing_area_set_dimension w=%d h=%d\n",w,h);
  w *= zoom_level;
  h *= zoom_level;
  gtk_widget_set_size_request(d, w, h);
  /* we absolutely need the scrollbars to get their correct value
   * right now */
  /* for now we do it this way */
  if (1){
    GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
    GtkAdjustment *wa = gtk_viewport_get_hadjustment(p);
    GtkAdjustment *ha = gtk_viewport_get_vadjustment(p);
    wa->upper = w;
    ha->upper = h;
  }
}

void gtk_wrapper_drawing_area_view(void *d, int x, int y, int w, int h,
                                   int zoom_level)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *wa = gtk_viewport_get_hadjustment(p);
  GtkAdjustment *ha = gtk_viewport_get_vadjustment(p);
  dprintf("gtk_wrapper_drawing_area_view x=%d y=%d w=%d h=%d\n",x,y,w,h);
  dprintf("gtk_wrapper_drawing_area_view h adjust=%d v adjust = %d\n",
          (int)wa->upper, (int)ha->upper);
  x *= zoom_level;
  y *= zoom_level;
  w *= zoom_level;
  h *= zoom_level;
  gtk_adjustment_clamp_page(wa, x, x+w-1);
  gtk_adjustment_clamp_page(ha, y, y+h-1);
}

unsigned int gtk_wrapper_add_audio_callback(int fd , void *fn, void *d)
{
  return gtk_input_add_full(fd, GDK_INPUT_WRITE, (GdkInputFunction)fn,
                            NULL, d, NULL);
}

void gtk_wrapper_remove_callback(unsigned int id)
{
  gtk_input_remove(id);
}

void gtk_set_color(int color)
{
  GdkGCValues values;
  if (color == current_color) return;
  current_color = color;
  values.foreground = colors[color];
  gdk_gc_set_values(gc, &values, GDK_GC_FOREGROUND);
}

/* returns the visible first y and last y of the drawing area */
void gtk_wrapper_get_min_max(void *d, int *m, int *M, int zoom_level)
{
  /* We may have flickering in the following case:
   * - the scrollbar is visible
   * - it hides a very long line
   * - only short lines are visible
   * - we thus return a min/max that will lead to the disparition of
   *   the scrollbar
   * - later on the scroll is removed
   * - the new (long) line is thus visible
   * - we come back here and give another M
   * - the new m/M stuff leads to the re-displaying of the scrollbar
   * - the scrollbar is back and the calculation restarts
   * so we get a flickering.
   * To avoid this situation we take the allocated height of the
   * container (which is a scrolled_window) which will always be
   * the same whatever the visibility of the scrollbar.
   */
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *ha = gtk_viewport_get_vadjustment(p);
  GtkWidget *pp = gtk_widget_get_parent(GTK_WIDGET(p));
  int container_height;

  container_height = pp->allocation.height;

  *m = ha->value / zoom_level;
  *M = (ha->value + container_height - 1) / zoom_level;
  dprintf("gtk_wrapper_get_min_max min=%d max=%d\n", *m, *M);
}

/* for get_width and get_height:
 * I don't know what to use between gtk_widget_get_size_request
 * and gdk_drawable_get_size, so I take maximum of both.
 * I also need to return partially visible pixels (when
 * zoom > 1 the division may "forget" some pixels, we
 * need the + zoom_level - 1 to count them)
 */
int gtk_wrapper_get_width(void *d, int zoom_level)
{
  int w, h;
  int w2, h2;
  gtk_widget_get_size_request(d, &w, &h);
  dprintf("gtk_wrapper_get_width w = %d h = %d\n", w, h);
  gdk_drawable_get_size(((GtkWidget *)d)->window, &w2, &h2);
  dprintf("gtk_wrapper_get_width 2 w2 = %d h2 = %d\n", w2, h2);
  return ((w > w2 ? w : w2) + zoom_level - 1)/ zoom_level;
}

int gtk_wrapper_get_height(void *d, int zoom_level)
{
  int w, h;
  int w2, h2;
  gtk_widget_get_size_request(d, &w, &h);
  dprintf("gtk_wrapper_get_height w = %d h = %d\n", w, h);
  gdk_drawable_get_size(((GtkWidget *)d)->window, &w2, &h2);
  dprintf("gtk_wrapper_get_height 2 w2 = %d h2 = %d\n", w2, h2);
  return ((h > h2 ? h : h2) + zoom_level - 1)/ zoom_level;
}

int gtk_wrapper_get_visible_height(void *d, int zoom_level)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *a = gtk_viewport_get_vadjustment(p);
  return a->page_size / zoom_level;
}

void *gtk_wrapper_get_tutorial_iter(void)
{
  /* ohhhhh!! a SINGLETON!! hell I spent some time in private companies,
   * this is the proof!
   */
  /* now let me justify this. gtk_wrapper.c is there so that we don't need
   * to include <gtk/gtk.h> when we want to use gtk, because compiling
   * witht this include takes too much time on my (who said "old"!?)
   * host. So we hide gtk in this file. And we need a GtkTextIter, so
   * what to do, huh? We could include the relevant stuff but heck,
   * it might break something in case gtk's people want to change
   * their includes. As a gtk's user we are only supposed to include
   * <gtk/gtk.h> so heck, we only play with void *, which means
   * to declare a GtkTextIter in this file and returning its address
   * as a void * to the rest of the app. (And writing a long useless
   * comment as a justification.)
   */
  static GtkTextIter iter;
  return &iter;
}

void gtk_wrapper_text_buffer_get_iter_at_offset(void *a, void *b, int c)
{
  gtk_text_buffer_get_iter_at_offset(a, b, c);
}

void gtk_wrapper_text_buffer_place_cursor(void *buffer, void *iter)
{
  gtk_text_buffer_place_cursor(buffer, iter);
}

void gtk_wrapper_insert_one(void *a, void *b, char *c, int d, char *e)
{
  gtk_text_buffer_insert_with_tags_by_name(a, b, c, d, e, NULL);
}

void gtk_wrapper_insert_two(void *a, void *b, char *c, int d, char *e,
                            char *f)
{
  gtk_text_buffer_insert_with_tags_by_name(a, b, c, d, e, f, NULL);
}

void gtk_wrapper_clear_text(void *b)
{
  gtk_text_buffer_set_text(b, "", -1);
}

void gtk_wrapper_text_view_begin(void *win)
{
  GtkAdjustment *wa = gtk_scrolled_window_get_hadjustment(win);
  GtkAdjustment *ha = gtk_scrolled_window_get_vadjustment(win);
  gtk_adjustment_clamp_page(wa, 0, 1);
  gtk_adjustment_clamp_page(ha, 0, 1);
}

void gtk_wrapper_string_size(void *p, int zoom, void *txt, int *width,
                             int *height)
{
  PangoLayout **layout = p;
  pango_layout_set_text(layout[zoom-1], txt, -1);
  pango_layout_get_pixel_size(layout[zoom-1], width, height);
  /* we have to divide by zoom level, so add pixels
   * so that later on the width and height will be overestimated
   * and not underestimated (because '(a+(b-1))/b' and then multiply
   * this result (which is rounded) by 'b' will give a number higher or
   * equal to 'a')
   * (yeah, ugly hack, but Pango may use any font so we have to
   * calculate the dimension of text using the font of the zoom level,
   * we cannot scale) (on my host when I scaled it was an overestimate,
   * on a newer one it was an underestimate)
   */
  *width += zoom - 1;
  *height += zoom - 1;
  *width /= zoom;
  *height /= zoom;
  dprintf("pango size width=%d height=%d\n", *width, *height);
}

void gtk_wrapper_draw_string(void *draw, void *p, char *text,
                             int x, int y,
                             int zoom)
{
  PangoLayout **layout = p;
  pango_layout_set_text(layout[zoom-1], text, -1);
  gdk_draw_layout(((GtkWidget *)draw)->window, gc,
                  x * zoom, y * zoom, layout[zoom-1]);
}

int gtk_wrapper_get_scroll_value_x(void *d)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *a = gtk_viewport_get_hadjustment(p);
  return gtk_adjustment_get_value(a);
}

int gtk_wrapper_get_scroll_value_y(void *d)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *a = gtk_viewport_get_vadjustment(p);
  return gtk_adjustment_get_value(a);
}

void gtk_wrapper_set_scroll_value_x(void *d, int x)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *a = gtk_viewport_get_hadjustment(p);
  /* don't ask me why but I must clip by myself (I have gtk 2.2.1) */
  if (x < a->lower) x = a->lower;
  /* not sure about the -1 here */
  if (x > a->upper - a->page_size-1) x = a->upper - a->page_size-1;
  dprintf("setting x=%d\n", x);
  gtk_adjustment_set_value(a, x);
}

void gtk_wrapper_set_scroll_value_y(void *d, int y)
{
  GtkViewport *p = (GtkViewport *)gtk_widget_get_parent(d);
  GtkAdjustment *a = gtk_viewport_get_vadjustment(p);
  /* don't ask me why but I must clip by myself (I have gtk 2.2.1) */
  if (y < a->lower) y = a->lower;
  /* not sure about the -1 here */
  if (y > a->upper - a->page_size-1) y = a->upper - a->page_size-1;
  dprintf("setting y=%d\n", y);
  gtk_adjustment_set_value(a, y);
}

void *gtk_wrapper_create_bitmap(char *data, int width, int height)
{
  return gdk_bitmap_create_from_data(gdk_get_default_root_window(),
                                     data, width, height);
}
