/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

/* note on history stuff:
 * for cgscore (might be found at http://sed.free.fr but cgscore
 * is more or less a dead project, so...) I wanted to have
 * multiple histories to have various sub-versions in one
 * document. I don't know if it's useful or not but it seemed
 * fun when I came to the idea. So in here (which is derived
 * from cgscore's code) you see "dad," "son" and "bro" stuff even
 * if in gcomposer we handle only one history's line.
 * (It's the LTL way of life as opposed to the CTL's one,
 * for the few of you who could understand this stuff.)
 * (Yeah I could have cleaned the shit up...)
 */

#include "history.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "doc.h"
#include "error.h"

/* At any time we can be anywhere in the history's line.
 * When the user adds an history, all the histories that
 * were previously undone are lost.
 * This function implements this loss.
 */
static void forget_future(doc *d)
{
  history *h = &d->history;
  history_item *cur;
  int i;

  if (h->current == -1) return;

  for (i = h->size-1; i > h->current; i--) {
    cur = &h->data[i];
    ERASE(d, cur->ht, (char *)h->history_data+cur->data);
    h->size--;
    h->history_size -= abs(cur->size);
  }

  cur = &h->data[h->current];
  cur->son = -1;
}

void destroy_history(doc *d)
{
  history *h = &d->history;
  history_item *cur;
  int i;

  for (i = h->size-1; i > 0; i--) {
    cur = &h->data[i];
    ERASE(d, cur->ht, (char *)h->history_data+cur->data);
    h->size--;
    h->history_size -= abs(cur->size);
  }

  if (h->maxsize) free(h->data);
  if (h->history_maxsize) free(h->history_data);

  INIT_HISTORY(d->history);
}

#ifdef DEBUG_ON
void dump_history(history *h, int item, int depth)
{
  int i;
  fprintf(stderr, "%3.3d", item);
  if (h->current == item) {
    fprintf(stderr, "->");
    for (i=0; i<depth-2; i++) fprintf(stderr, " ");
  } else
    for (i=0; i<depth; i++) fprintf(stderr, " ");
  fprintf(stderr, "hist data=%d size=%d history=%d dad=%d son=%d bro=%d\n",
    h->data[item].data,
    h->data[item].size,
    h->data[item].ht,
    h->data[item].dad,
    h->data[item].son,
    h->data[item].bro);
  if (h->data[item].son != -1) dump_history(h, h->data[item].son, depth+2);
  if (h->data[item].bro != -1) dump_history(h, h->data[item].bro, depth);
}
#endif /* DEBUG_ON */

/* return <0 if error, 0 if ok */
int add_history(doc *d, history_type ht, void *data, int size)
{
  history *h = &d->history;
  history_item *i;
  history_item *dad;

  forget_future(d);

  /* size can't be 0 for major handling */
  if (size == 0) size = 1;
  if (h->size == h->maxsize) {
    history_item *new;
    if (h->maxsize)
      new = realloc(h->data, (h->maxsize+2048) * sizeof(history_item));
    else
      new = malloc(2048 * sizeof(history_item));
    if (!new) return -OOM;
    h->data = new;
    h->maxsize += 2048;
    /* node 0 always exists, it's everyone's dad (call it God) */
    if (h->size == 0) {
      h->size = 1;
      h->data[0].size = 0;
      h->data[0].data = 0;
      h->data[0].dad = -1;
      h->data[0].son = -1;
      h->data[0].bro = -1;
      h->current = 0;
    }
  }
  if (h->history_size + size > h->history_maxsize) {
    int new_size = h->history_maxsize + 1024;
    void *new;
    if (new_size < h->history_size + size)
      new_size = h->history_size + size;
    if (h->history_maxsize)
      new = realloc(h->history_data, new_size);
    else
      new = malloc(new_size);
    if (!new) return -OOM;
    h->history_data = new;
    h->history_maxsize = new_size;
  }
  memcpy((char *)h->history_data+h->history_size, data, size);
  i = &h->data[h->size];
  h->size++;
  i->data = h->history_size;
  h->history_size += size;
  i->size = size;
  if (h->next_is_major) {
    i->size = -size;
    h->next_is_major = 0;
  }
  i->ht = ht;
  i->dad = h->current;
  i->son = -1;
  i->bro = -1;
  h->current = h->size-1;
  /* the dad's got a new son */
  dad = &h->data[i->dad];
  i->bro = dad->son;
  dad->son = h->size-1;
  return 0;
}

void history_up(doc *d)
{
  history *h = &d->history;
  history_item *i;
  if (h->current == -1) return;
  while (h->current != 0 && h->data[h->current].size > 0) {
    i = &h->data[h->current];
    h->current = i->dad;
    UNDO(d, i->ht, (char *)h->history_data+i->data);
  }
  if (h->current == 0) return;
  i = &h->data[h->current];
  h->current = i->dad;
  UNDO(d, i->ht, (char *)h->history_data+i->data);
}

void history_down(doc *d, int son_number)
{
  history *h = &d->history;
  history_item *i;
  int cur;
  if (h->current == -1) return;
  i = &h->data[h->current];
  cur = i->son;
  while (cur != -1 && son_number) {
    cur = h->data[cur].bro;
    son_number--;
  }
  if (cur == -1) return;
  i = &h->data[cur];
  h->current = cur;
  REDO(d, i->ht, (char *)h->history_data+i->data);
  while (h->data[h->current].son != -1 &&
         h->data[h->data[h->current].son].size > 0) {
    h->current = h->data[h->current].son;
    i = &h->data[h->current];
    REDO(d, i->ht, (char *)h->history_data+i->data);
  }
}

void history_up_and_forget(doc *d)
{
  history *h = &d->history;
  int last_is_major = 0;
  history_item *i;
  history_item *dad;
  /* if next_is_major is set, we didn't add anything, so simply return */
  if (h->next_is_major) {
    h->next_is_major = 0;
    return;
  }
  /* we have to remove all histories up and including to a major point */
  while (!last_is_major) {
    if (h->current == -1 || h->current == 0) return;
    i = &h->data[h->current];
    last_is_major = i->size < 0;
    h->current = i->dad;
    UNDO(d, i->ht, (char *)h->history_data+i->data);
    ERASE(d, i->ht, (char *)h->history_data+i->data);
    h->size--;
    h->history_size -= abs(i->size);
    /* the dad lost a son */
    dad = &h->data[i->dad];
    dad->son = i->bro;
  }
}

/* <0 error, 0 ok */
int set_history_size(history *h, int size)
{
  if (size == 0) return 0;
  h->data = malloc(((size + 255) & ~255) * sizeof(history_item));
  if (!h->data) return -OOM;
  h->maxsize = (size + 255) & ~255;
  h->size = size;
  return 0;
}

/* <0 error, >=0 ok
 * and returned value is elem->data value
 */
int increase_history_data(history *h, int size, int *object_size)
{
  int ret;
  size += 3;
  size &= ~3;
  if (size == 0) size = 4;
  if (h->history_size + size > h->history_maxsize) {
    int new_size = h->history_maxsize + 1024;
    void *new;
    if (new_size < h->history_size + size)
      new_size = h->history_size + size;
    new = realloc(h->history_data, new_size);
    if (!new) return -OOM;
    h->history_data = new;
    h->history_maxsize = new_size;
  }
  ret = h->history_size;
  h->history_size += size;
  *object_size = size;
  return ret;
}
