/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#ifndef _HISTORY_H_
#define _HISTORY_H_

#include "typedefs.h"
#include "history_enum.h"

typedef struct {
  int data;	/* offset in the history_data array */
  int size;	/* can be < 0 in case of major point */
  history_type ht;
  int dad;	/* offset in the data array */
  int son;	/* offset in the data array */
  int bro;	/* offset in the data array */
} history_item;

typedef struct history {
  history_item *data;
  int size;
  int maxsize;
  void *history_data;
  int history_size;
  int history_maxsize;
  int current;
  int next_is_major;
} history;

#define INIT_HISTORY(h)		\
  do {				\
    h.data = 0;			\
    h.size = 0;			\
    h.maxsize = 0;		\
    h.history_data = 0;		\
    h.history_size = 0;		\
    h.history_maxsize = 0;	\
    h.current = -1;		\
    h.next_is_major = 0;	\
  } while (0)

#define FREE_HISTORY(h)		\
  do {				\
    free(h.data);		\
    free(h.history_data);	\
  } while (0)

void destroy_history(struct doc *);

int add_history(struct doc *, history_type, void *, int);
void history_up(struct doc *);
void history_down(struct doc *, int);
void history_up_and_forget(struct doc *);

int set_history_size(history *, int);
int increase_history_data(history *, int, int *);

#define history_next_is_major(d) (d)->next_is_major = 1

#define history_get_pos(h) (h)->current
#define history_get_len(h) (h)->size

#endif /* _HISTORY_H_ */
