/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

/* info line management */

#include "info.h"

#include "gtk_wrapper.h"
#include "dprintf.h"

#include <stdio.h>
#include <stdarg.h>

static char buf[2048];

void set_info(doc *d, char *s, ...)
{
  /* not sure about what vsnprintf does, so let's give 2047 and
   * then force 0 for buf[2047]
   */
  va_list va;
  va_start(va, s);
  dprintf("set_info (%s)\n", s);
  vsnprintf(buf, 2047, s, va);
  va_end(va);
  buf[2047] = 0;
  gtk_wrapper_label_set_text(d->desc_label, buf);
}

static int put(int c, int pos)
{
  buf[pos] = c;
  return pos == 2047 ? pos : pos + 1;
}

#if 0
static int put_number(int n, int pos)
{
  char _s[64];
  char *s = _s;
  sprintf(s, "%d", n);
  while (*s) { pos = put(*s, pos); s++; }
  return pos;
}
#endif

static int put_string(char *s, int pos)
{
  while (*s) { pos = put(*s, pos); s++; }
  return pos;
}

void set_info_chord(doc *d)
{
  int pos;
  int i;
  cursor *c = &d->cursor;
  dprintf("set_info_chord\n");
  if (c->line >= d->size ||
      LINE_TYPE(d, c->line) != LINE_MUSIC ||
      c->chord >= LINE_MUSIC_SIZE(d, c->line) ||
      CHORD_TYPE(d, c->line, c->chord) != NOTES_CHORD) {
    gtk_wrapper_label_set_text(d->desc_label, "");
    return;
  }

  pos = 0;

  for (i = 0; i < CHORD_SIZE(d, c->line, c->chord); i++) {
    elem *e = &CHORD_DATA(d, c->line, c->chord)[i];

    if (c->type == SMALL_CURSOR) {
      if (c->note == i) pos = put_string(" | ", pos);
    }

    switch (e->type) {
      case NOTE: {
        char *t;
        /* put pitch */
        int pitch = e->u.note.pitch - get_delta_clef(d, c->line, c->chord);
        while (pitch < 0) pitch += 7;
        pitch %= 7;
        pitch--;
        if (pitch == -1) pitch = 6;
        pos = put('A' + pitch, pos);
        /* put alteration */
        switch (e->u.note.alteration) {
          case NORMAL: break;
          case DIESE: pos = put('#', pos); break;
          case BEMOL: pos = put('b', pos); break;
          case BECARRE: pos = put_string("becarre", pos); break;
        }
        /* duration */
        pos = put('(', pos);
        switch (e->u.note.duration) {
          case RONDE: t = "1"; break;
          case BLANCHE: t = "1/2"; break;
          case NOIRE: t = "1/4"; break;
          case CROCHE: t = "1/8"; break;
          case DCROCHE: t = "1/16"; break;
          case DDCROCHE: t = "1/32"; break;
          case DDDCROCHE: t = "1/64"; break;
          case DDDDCROCHE: t = "1/128"; break;
        }
        pos = put_string(t, pos);
        /* delta stuff */
        if (e->u.note.head_delta) {
          char t[64];
          sprintf(t, " l:%d", e->u.note.head_delta);
          pos = put_string(t, pos);
        }
        if (e->u.note.alteration_delta) {
          char t[64];
          sprintf(t, " m:%d", e->u.note.alteration_delta);
          pos = put_string(t, pos);
        }
        if (e->u.note.dots_delta) {
          char t[64];
          sprintf(t, " o:%d", e->u.note.dots_delta);
          pos = put_string(t, pos);
        }
        if (e->u.note.stem_down) pos = put_string(" j", pos);
        pos = put(')', pos);
        /* nb dots */
        /* admire the fall-through */
        switch (e->u.note.nb_dots) {
          case 3: pos = put('.', pos);
          case 2: pos = put('.', pos);
          case 1: pos = put('.', pos);
        }
        break;
      }
      case SILENCE: {
        char *t;
        pos = put_string("rest", pos);
        /* duration */
        switch (e->u.silence.duration) {
          case RONDE: t = "(1)"; break;
          case BLANCHE: t = "(1/2)"; break;
          case NOIRE: t = "(1/4)"; break;
          case CROCHE: t = "(1/8)"; break;
          case DCROCHE: t = "(1/16)"; break;
          case DDCROCHE: t = "(1/32)"; break;
          case DDDCROCHE: t = "(1/64)"; break;
          case DDDDCROCHE: t = "(1/128)"; break;
        }
        pos = put_string(t, pos);
        /* nb dots */
        /* admire the fall-through */
        switch (e->u.silence.nb_dots) {
          case 3: pos = put('.', pos);
          case 2: pos = put('.', pos);
          case 1: pos = put('.', pos);
        }
        break;
      }
    }
    if (i != CHORD_SIZE(d, c->line, c->chord)-1)
      pos = put_string(", ", pos);
  }

  if (pos == 2047) pos--;
  (void)put(0, pos);

  gtk_wrapper_label_set_text(d->desc_label, buf);
}
