/*
 * m: a language for cgscore to describe states
 * public domain, like cgscore
 */

#include "c_fun_args.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "list.h"
#include "string.h"
#include "debug.h"

static char stack[128];
static int stack_size;

/* -1 error 0 ok */
int push_or_pop(char s)
{
  switch(s) {
    case ')' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '(') goto pop_error;
      break;
    case ']' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '[') goto pop_error;
      break;
    case '}' :
      if (stack_size == 0) goto pop_error;
      stack_size--;
      if (stack[stack_size] != '{') goto pop_error;
      break;
    default:
      if (s != '(' && s != '[' && s != '{') return 0;
      if (stack_size == 128) {
        fprintf(stderr, "parenthesis too much nested\n");
        fprintf(stderr, "contact the autors for them to fix the bug\n");
        return -1;
      }
      stack[stack_size] = s;
      stack_size++;
      break;
  }
  return 0;

pop_error:
  fprintf(stderr, "wrong parenthesis\n");
  return -1;
}

/* sp is head of string to parse and does never move */
static char *sp;
/* p moves */
static char *p;

void remove_white_spaces(void)
{
  while (*p && isspace(*p)) p++;
}

/* -1 error 0 ok */
int parse_struct_union_enum(void)
{
  int has_identifier = 0;

  remove_white_spaces();
  if (!*p) return -1;

  /* optional identifier */
  if (*p == '_' || isalpha(*p)) {
    has_identifier = 1;
    while (*p && (*p == '_' || isalnum(*p))) p++;
  }

  remove_white_spaces();
  if (!*p) return -1;

  if (*p != '{' && !has_identifier) goto sue_error;

  if (push_or_pop(*p) == -1) return -1;
  while (*p && stack_size) {
    p++;
    if (push_or_pop(*p) == -1) return -1;
  }
  if (!*p) goto sue_error;
  p++;

  return 0;

sue_error:
  fprintf(stderr, "bad union/struct/enum declaration\n");
  return -1;
}

/* -1 error, 0 ok and variables set correctly */
int parse_argument(int *start_offset, int *end_offset,
                   int *start_name, int *end_name)
{
  *start_offset = p-sp;

  /* we accept more than what the spec (c99) says, but anyway the C compiler
   * will complain later on, so no problem
   */
  int type_specified = 0;

  /* first stuff: type specifier */
  /* barbarian code */
  while (1) {
    SLOW;
    remove_white_spaces();
    if (!*p) goto arg_error;

    if (!strncmp(p, "typedef",  7)) { p+=7; continue; }
    if (!strncmp(p, "extern",   6)) { p+=6; continue; }
    if (!strncmp(p, "static",   6)) { p+=6; continue; }
    if (!strncmp(p, "auto",     4)) { p+=4; continue; }
    if (!strncmp(p, "register", 8)) { p+=8; continue; }

    if (!strncmp(p, "const",    5)) { p+=5; continue; }
    if (!strncmp(p, "restrict", 8)) { p+=8; continue; }
    if (!strncmp(p, "volatile", 8)) { p+=8; continue; }

    if (!strncmp(p, "inline",   6)) { p+=6; continue; }

    if (!strncmp(p, "void",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "char",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "short",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "int",      3)) { p+=3; type_specified = 1; continue; }
    if (!strncmp(p, "long",     4)) { p+=4; type_specified = 1; continue; }
    if (!strncmp(p, "float",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "double",   6)) { p+=6; type_specified = 1; continue; }
    if (!strncmp(p, "signed",   6)) { p+=6; type_specified = 1; continue; }
    if (!strncmp(p, "unsigned", 8)) { p+=8; type_specified = 1; continue; }
    if (!strncmp(p, "_Bool",    5)) { p+=5; type_specified = 1; continue; }
    if (!strncmp(p, "_Complex", 8)) { p+=8; type_specified = 1; continue; }
    if (!strncmp(p, "_Imaginary", 10)) { p+=10; type_specified = 1; continue; }

    if (!strncmp(p, "struct", 6) || !strncmp(p, "union", 5) ||
        !strncmp(p, "enum", 4)) {
      if (*p == 's') p+=6;
      else if (*p == 'u') p+=5;
      else p += 4;
      if (parse_struct_union_enum() == -1) return -1;
      type_specified = 1;
      continue;
    }

    /* might be a user defined type name */
    if (*p == '_' || isalpha(*p)) {
      /* if type_specified is true this identifier must be understood as the
       * name of the argument
       */
      if (type_specified) break;
      while (*p && (*p == '_' || isalnum(*p))) p++;
      type_specified = 1;
      continue;
    }

    /* might be the beginning of a declarator */
    if (*p == '*' || *p == '(') break;

    goto arg_error;
  }

  /* now: declarator */
  while (1) {
    remove_white_spaces();
    if (!*p) goto arg_error;
    if (push_or_pop(*p) == -1) return -1;
    if (*p == '_' || isalpha(*p)) break;
    if (*p != '*' && *p !='(') goto arg_error;
    p++;
  }

  /* here is the identifier */
  *start_name = p-sp;
  while (*p && (*p == '_' || isalnum(*p))) p++;
  *end_name = p-1-sp;

  /* now, let's eat up to the end of the declaration */
  while (*p && (stack_size || *p != ',')) {
    if (push_or_pop(*p) == -1) return -1;
    p++;
  }
  if (stack_size) goto arg_error;

  *end_offset = p-1-sp;

  if (*p) p++;

  return 0;

arg_error:
  fprintf(stderr, "bad arguments' declaration\n");
  return -1;
}

/* -1 error, list if ok
 * structure of list is:
 * <arguments string> <argument 1 start offset> <argument 1 end offset>
 *   <argument 1 name start offset> <argument 1 name end offset>
 *   <argument m start offset> <argument m end offset>
 *   <argument m name start offset> <argument m name end offset>
 */
list parse_arguments(string s)
{
  int start_offset;
  int end_offset;
  int start_name;
  int end_name;
  list l = new_list();
  if (l == -1) return -1;

  if (list_add_int(l, s) == -1) return -1;

  p = get_string(s);
  sp = p;

  remove_white_spaces();
  while (*p) {
    if (parse_argument(&start_offset, &end_offset,
                       &start_name, &end_name) == -1) return -1;
    if (list_add_int(l, start_offset) == -1) return -1;
    if (list_add_int(l, end_offset) == -1) return -1;
    if (list_add_int(l, start_name) == -1) return -1;
    if (list_add_int(l, end_name) == -1) return -1;
    remove_white_spaces();
  }

  return l;
}

/* 0 error, p otherwise with removing of leading comments (if there are) */
static char *remove_comments(char *p)
{
  if (*p != '/') return p;
  p++;

  if (*p == '*') {
    while (1) {
      do {
        p++;
        if (*p == 0) { fprintf(stderr, "bad C comment\n"); return 0; }
      } while (*p != '*');
      p++;
      if (*p == '/') return p+1;
    }
  }

  if (*p == '/') {
    while (1) {
      p++;
      if (*p == '\n') return p+1;
      if (*p == 0) { fprintf(stderr, "bad C comment\n"); return 0; }
    }
  }

  return p;
}

/* 0 error, p without the starting string if ok */
static char *remove_string(char *p)
{
  char sep = *p;
  int last_escape = 0;

  do {
    if (last_escape) last_escape = 0; else last_escape = (*p == '\\');
    p++;
    if (*p == 0) {
      fprintf(stderr, "unterminated string sequence\n");
      return 0;
    }
  } while (*p != sep || last_escape);

  return p+1;
}

/* 0 error, pointer of ';' in p (not in comment, not in string) if ok */
static char *to_dot_comma(char *p)
{
  int depth = 0;

  while (*p != ';' || depth) {
    p = remove_comments(p);
    if (!p) return 0;

    if (*p == '(') depth++;
    if (*p == ')') depth--;
    if (*p == '"' || *p == '\'') {
      p = remove_string(p);
      if (!p) return 0;
      p--;
    }
    p++;
    if (*p == 0) {
      fprintf(stderr, "bad CODE (forgot the ';'? bad parenthesis?");
      fprintf(stderr, " bad string?)\n");
      return 0;
    }
  }

  return p;
}

/* -1 if error, a list if ok. Structure of list:
 *   <code string> <call 1 start offset> <call 1 end offset>
 *     <call m start offset> <call m end offset>
 */
list parse_code(string code)
{
  char *s = get_string(code);
  char *p = s;
  list l = new_list();

  if (l == -1) return -1;

  if (list_add_int(l, code) == -1) return -1;

  while (1) {
    char *q;
    q = strstr(p, "CALL(");
    if (!q) break;
    if (list_add_int(l, q-s) == -1) return -1;
    q = to_dot_comma(q);
    if (!q) return -1;
    if (list_add_int(l, q-s) == -1) return -1;
    p = q+1;
  }

  return l;
}
