/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "parachute.h"

#ifdef PARACHUTE_ON

#include <X11/Xlib.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#endif /* PARACHUTE_ON */

#include "file.h"
#include "doc.h"
#include "error.h"

doc *parachute_doc;

#ifdef PARACHUTE_ON

static int sig;

static void do_abort(void)
{
  int i;
  /* re-default signals not to get a loop that will lead to stack overflow
   * (it would core dump anyway but it's not clean)
   */
  for (i = 0; i < 101; i++) signal(i, SIG_DFL);
  fprintf(stderr, "Aborting.\n");
  abort();
}

static void _parachute(void)
{
  char *t;
  char s[FILENAME_MAX];
  static int in_parachute = 0;

  if (!parachute_doc) {
nodoc:
    if (sig != SIGINT) {
      fprintf(stderr,
              "The program has encountered a fatal error. It has to quit.\n");
      fprintf(stderr,
              "Contact the authors for them to correct the problem.\n");
    }
    fprintf(stderr, "Your document (if any) is up to date, no need to ");
    fprintf(stderr, "create a backup.\n");
    exit(1);
  }

  /* don't save if not modified */
  if (parachute_doc->state != STATE_MODIFIED) goto nodoc;

  if (in_parachute) {
    /* huho, big trouble here, let's be barbarian */
    fprintf(stderr, "The rescue has failed.\n");
    fprintf(stderr, "The program will now abort for you to (maybe) have a\n");
    fprintf(stderr, "core file that you can use to recover some data or\n");
    fprintf(stderr, "analyze to understand why the program failed.\n");
    fprintf(stderr,
            "Please, contact the authors for them to improve the program.\n");
    do_abort();
  }
  in_parachute = 1;

  if (sig != SIGINT) {
    fprintf(stderr,
            "The program has encountered a fatal error. It has to quit.\n");
    fprintf(stderr, "Contact the authors for them to correct the problem.\n");
  } else {
    fprintf(stderr, "You requested to quit the program but your document ");
    fprintf(stderr, "is modified.\n");
  }
  fprintf(stderr, "The program is now trying to save the document.\n");

  /* try to save $HOME/gcomposer.rescue */
  t = getenv("HOME");
  if (t && strlen(t) + strlen("/gcomposer.rescue") + 1 <= FILENAME_MAX) {
    sprintf(s, "%s/gcomposer.rescue", t);
    fprintf(stderr, "Trying to save to %s\n", s);
    if (save(parachute_doc, s)) goto save_success;
    fprintf(stderr, "Saving failed.\n");
  }

  /* failed? try $TEMP/gcomposer.rescue */
  t = getenv("TEMP");
  if (t && strlen(t) + strlen("/gcomposer.rescue") + 1 <= FILENAME_MAX) {
    sprintf(s, "%s/gcomposer.rescue", t);
    fprintf(stderr, "Trying to save to %s\n", s);
    if (save(parachute_doc, s)) goto save_success;
    fprintf(stderr, "Saving failed.\n");
  }

  /* failed? try $TMP/gcomposer.rescue */
  t = getenv("TMP");
  if (t && strlen(t) + strlen("/gcomposer.rescue") + 1 <= FILENAME_MAX) {
    sprintf(s, "%s/gcomposer.rescue", t);
    fprintf(stderr, "Trying to save to %s\n", s);
    if (save(parachute_doc, s)) goto save_success;
    fprintf(stderr, "Saving failed.\n");
  }

  /* failed? try /tmp/gcomposer.rescue */
  fprintf(stderr, "Trying to save to /tmp/gcomposer.rescue\n");
  if (save(parachute_doc, "/tmp/gcomposer.rescue")) goto save_success;
  fprintf(stderr, "Saving failed.\n");

  /* failed? hell, what to do now? */
  fprintf(stderr, "The program failed saving your document.\n");
  fprintf(stderr, "The program will now abort for you to (maybe) have a\n");
  fprintf(stderr, "core file that you can use to recover some data or\n");
  fprintf(stderr, "analyze to understand why the program failed.\n");
  fprintf(stderr,
          "Please, contact the authors for them to improve the program.\n");
  do_abort();

save_success:
  fprintf(stderr, "The file has been saved successfully.\n");
  fprintf(stderr, "Since it is a rescue save, it might not be correct.\n");
  if (sig != SIGINT)
    fprintf(stderr,
          "Please, contact the authors for them to improve the program.\n");
  exit(1);
}

static void _parachute_sig(int s)
{
  sig = s;
  fprintf(stderr, "[Signal %d caught.]\n", s);
  _parachute();
}

static void _parachute_x(void)
{
  sig = -1;
  fprintf(stderr, "[X Window failure.]\n");
  _parachute();
}

#endif /* PARACHUTE_ON */

/* -1 error, 0 ok */
int parachute(void)
{
#ifndef PARACHUTE_ON
  return 0;
#else

  /* protect against all the signals that would quit the program
   * (except SIGKILL and SIGSTOP that can't be caught)
   */
  /* ugly code, huh? */
#define SIG(x) if (signal(x, _parachute_sig) == SIG_ERR) return -1
#ifdef SIGHUP
  SIG(SIGHUP);
#endif
#ifdef SIGINT
  SIG(SIGINT);
#endif
#ifdef SIGQUIT
  SIG(SIGQUIT);
#endif
#ifdef SIGILL
  SIG(SIGILL);
#endif
#ifdef SIGABRT
  SIG(SIGABRT);
#endif
#ifdef SIGFPE
  SIG(SIGFPE);
#endif
#ifdef SIGSEGV
  SIG(SIGSEGV);
#endif
#ifdef SIGPIPE
  SIG(SIGPIPE);
#endif
#ifdef SIGALRM
  SIG(SIGALRM);
#endif
#ifdef SIGTERM
  SIG(SIGTERM);
#endif
#ifdef SIGUSR1
  SIG(SIGUSR1);
#endif
#ifdef SIGUSR2
  SIG(SIGUSR2);
#endif
#ifdef SIGBUS
  SIG(SIGBUS);
#endif
#ifdef SIGPOLL
  SIG(SIGPOLL);
#endif
#ifdef SIGPROF
  SIG(SIGPROF);
#endif
#ifdef SIGSYS
  SIG(SIGSYS);
#endif
#ifdef SIGTRAP
  SIG(SIGTRAP);
#endif
#ifdef SIGVTALRM
  SIG(SIGVTALRM);
#endif
#ifdef SIGXCPU
  SIG(SIGXCPU);
#endif
#ifdef SIGXFSZ
  SIG(SIGXFSZ);
#endif
#ifdef SIGIOT
  SIG(SIGIOT);
#endif
#ifdef SIGEMT
  SIG(SIGEMT);
#endif
#ifdef SIGSTKFLT
  SIG(SIGSTKFLT);
#endif
#ifdef SIGIO
  SIG(SIGIO);
#endif
#ifdef SIGPWR
  SIG(SIGPWR);
#endif
#ifdef SIGLOST
  SIG(SIGLOST);
#endif
#ifdef SIGUNUSED
  SIG(SIGUNUSED);
#endif
  /* realtime signals? let's not care for the moment */

  /* The XLib also likes to kill its clients */
  XSetIOErrorHandler((int (*)(Display *))_parachute_x);

  return 0;
#endif /* PARACHUTE_ON */
}
