/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

/* there is no clear_cursor in there because invalidate_line does the
 * trick (ok there is one for the tab case, but you got the point)
 * if we ever change the way it works we need to not forget the clear_cursor
 * stuff
 */

state state_selection(doc *d, gui *g, int key, int *ret_code)
{
  data {
    doc *d;
    int p1_line;
    int p1_chord;
    int p2_line;
    int p2_chord;
    gui *g;
    int *ret_code;
    char load_file[FILENAME_MAX];
  }

  enter {
    data->d = d;
    data->g = g;
    data->ret_code = ret_code;

    data->p1_line = data->d->cursor.line;
    data->p1_chord = data->d->cursor.chord;
    data->p2_line = data->d->cursor.line;
    data->p2_chord = data->d->cursor.chord;

    switch (key) {
      default: /* normally you should not come here */ break;
      case GDK_Right:
        data->d->cursor.chord++;
        data->p2_chord++;
        break;
      case GDK_Left:
        data->d->cursor.chord--;
        data->p2_chord--;
        break;
      case GDK_Up:
        invalidate_line(data->d, data->p1_line);
        data->d->cursor.line--;
        data->d->cursor.chord = 0;
        data->p2_line--;
        data->p2_chord = 0;
        break;
      case GDK_Down:
        invalidate_line(data->d, data->p1_line);
        data->d->cursor.line++;
        data->d->cursor.chord = 0;
        data->p2_line++;
        data->p2_chord = 0;
        break;
    }

    if (is_after(data->p2_line, data->p2_chord,
                 data->p1_line, data->p1_chord)) {
      data->d->sel.start_line = data->p1_line;
      data->d->sel.start_chord = data->p1_chord;
      data->d->sel.end_line = data->p2_line;
      data->d->sel.end_chord = data->p2_chord;
    } else {
      data->d->sel.start_line = data->p2_line;
      data->d->sel.start_chord = data->p2_chord;
      data->d->sel.end_line = data->p1_line;
      data->d->sel.end_chord = data->p1_chord;
    }

    invalidate_line(data->d, data->p2_line);

    *data->ret_code = I_NOTHING;

    set_info(data->d, "Selection's mode (press ESC to leave it)");

    /* no paste available */
    grey(GREY_MENU_PASTE);

    /* copy available */
    ungrey(GREY_MENU_COPY);
  }

  leave {
    data->d->sel.start_line = -1;
    /* a bit barbarian, huh? */
    invalidate_all(data->d);
    set_info(data->d, "");
    ungrey(GREY_MENU_PASTE);
    grey(GREY_MENU_COPY);
  }

  action key_press(int key, int control, int shift) {
    /* we must not return if the user presses control or shift or ... */
    /* This list has been made based on my keyboard (pc102 or something)
     * and may not be correct. Hell, it's just a hack.
     */
    if (key == GDK_Control_L || key == GDK_Control_R ||
        key == GDK_Shift_L || key == GDK_Shift_R ||
        key == GDK_Caps_Lock || key == GDK_Shift_Lock ||
        key == GDK_Alt_L || key == GDK_Alt_R ||
        key == GDK_Meta_L || key == GDK_Meta_R ||
        key == GDK_Super_L || key == GDK_Super_R ||
        key == GDK_Hyper_L || key == GDK_Hyper_R ||
        key == GDK_Menu || key == GDK_Multi_key)
      goto end_switch;

    if (control) {
      switch (key) {
        case 's':
          DO_SAVE(_owner);
          /* after a DO_XX, we must return 0 */
          return 0;
        case 'z':
          DO_UNDO(_owner);
          return 0;
        case 'y':
          DO_REDO(_owner);
          return 0;
        case 'c': /* fall through */
        case GDK_Insert: goto case_copy;
      }
      set_info(data->d, "Selection's mode (press ESC to leave it)");
      goto end_switch;
    }

    /* TAB or shift TAB to change the select point */
    if (key == GDK_Tab) {
      int dummy;
      clear_cursor(&data->d->cursor);
      data->d->cursor.chord = data->p1_chord;
      data->d->cursor.line = data->p1_line;
      dummy = data->p1_chord;
      data->p1_chord = data->p2_chord;
      data->p2_chord = dummy;
      dummy = data->p1_line;
      data->p1_line = data->p2_line;
      data->p2_line = dummy;
      goto end_switch;
    }

    /* we want to do nothing in this state for delete/back-delete */
    if (key == GDK_Delete || key == GDK_BackSpace) {
      set_info(data->d, "Selection's mode (press ESC to leave it)");
      goto end_switch;
    }

    /* leave the selection state by pressing escape */
    if (key == GDK_Escape) RETURN;

    if (!shift) {
      set_info(data->d, "Selection's mode (press ESC to leave it)");
      goto end_switch;
    }

    switch (key) {
      default:
        set_info(data->d, "Selection's mode (press ESC to leave it)");
        goto end_switch;
      case GDK_Right:
        /* if end of line, can't go right */
        switch (LINE_TYPE(data->d, data->d->cursor.line)) {
          case LINE_MUSIC:
            if (data->d->cursor.chord ==
                data->d->data[data->d->cursor.line].u.music.size)
              goto end_switch;
            break;
          case LINE_TEXT:
            /* only 2 positions for text lines */
            if (data->d->cursor.chord == 1)
              goto end_switch;
            break;
        }
        data->d->cursor.chord++;
        data->p2_chord++;
        break;
      case GDK_Left:
        if (data->d->cursor.chord == 0)
          goto end_switch;
        data->d->cursor.chord--;
        data->p2_chord--;
        break;
      case GDK_Up:
        if (data->d->cursor.line == 0)
          goto end_switch;
        invalidate_line(data->d, data->p2_line);
        data->d->cursor.line--;
        data->d->cursor.chord = 0;
        data->p2_line--;
        data->p2_chord = 0;
        break;
      case GDK_Down:
        if (data->d->cursor.line == data->d->size - 1)
          goto end_switch;
        invalidate_line(data->d, data->p2_line);
        data->d->cursor.line++;
        data->d->cursor.chord = 0;
        data->p2_line++;
        data->p2_chord = 0;
        break;
    }

    if (is_after(data->p2_line, data->p2_chord,
                 data->p1_line, data->p1_chord)) {
      data->d->sel.start_line = data->p1_line;
      data->d->sel.start_chord = data->p1_chord;
      data->d->sel.end_line = data->p2_line;
      data->d->sel.end_chord = data->p2_chord;
    } else {
      data->d->sel.start_line = data->p2_line;
      data->d->sel.start_chord = data->p2_chord;
      data->d->sel.end_line = data->p1_line;
      data->d->sel.end_chord = data->p1_chord;
    }

    invalidate_line(data->d, data->p2_line);
    goto end_switch;

case_copy:
    if (!copy_selection(data->d)) {
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    set_info(data->d, "Selection successfully copied");

    /* the paste menu becomes available */
    forcegrey(GREY_MENU_PASTE, 1);

    goto end_switch;

end_switch:
    reset_cursor(&data->d->cursor);
  }

  action selection_copy {
    if (!copy_selection(data->d))
      CALL(error_state, data->g->win);
    set_info(data->d, "Selection successfully copied");

    /* the paste menu becomes available */
    forcegrey(GREY_MENU_PASTE, 1);
  }

%%import "states/common_actions.i"
}
