/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

state t_state(doc *d, gui *g, int *key, int *control, int *shift,
              int *ret_code)
{
  data {
    doc *d;
    struct gui *g;
    int *key;
    int *control;
    int *shift;
    int *ret_code;
    char load_file[FILENAME_MAX];
  }

  enter {
    data->d = d;
    data->g = g;
    data->ret_code = ret_code;
    data->key = key;
    data->control = control;
    data->shift = shift;
    *data->ret_code = I_NOTHING;
  }

  leave {
    /* do nothing */
  }

  action key_press(int key, int control, int shift) {
    int pitch_direction;

    if (control) goto not_handled_key;

    switch (key) {
      default: goto not_handled_key;
      case '#': goto case_change_alteration;
      case 'u': pitch_direction = 1; goto case_change_position;
      case 'U': pitch_direction = -1; goto case_change_position;
    }

case_change_alteration: {
    int alteration = data->d->data[data->d->cursor.line].u.music.
                              data[data->d->cursor.chord].
                              u.tonalite.alteration;

    switch (alteration) {
      case DIESE: alteration = BEMOL; break;
      case BEMOL: alteration = BECARRE; break;
      case BECARRE: alteration = DIESE; break;
      case NORMAL: /* we can't come here */ goto end_switch;
    }
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_tonalite_alteration(data->d, data->d->cursor.line,
                                data->d->cursor.chord, alteration) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    goto end_switch;
  }

case_change_position: {
    int position = data->d->data[data->d->cursor.line].u.music.
                         data[data->d->cursor.chord].u.tonalite.position;

    position += pitch_direction;
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (set_tonalite_position(data->d, data->d->cursor.line,
                              data->d->cursor.chord, position) < 0) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      goto set_error;
    }
    size_post_treatment(data->d);
    goto end_switch;
  }

not_handled_key:
    *data->key = key;
    *data->control = control;
    *data->shift = shift;
    *data->ret_code = I_UNSUPPORTED_KEY;
    RETURN;

set_error:
    CALL(error_state, data->g->win);
    goto end_switch;

end_switch:
    reset_cursor(&data->d->cursor);

    /* set the grey for undo/redo */
    set_grey_undo_redo(data->d);
  }

%%import "states/common_actions.i"
}
