/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

state text_state(doc *d, gui *g, int *ret_code)
{
  data {
    doc *d;
    struct gui *g;
    int *ret_code;
    char load_file[FILENAME_MAX];
  }

  enter {
    data->d = d;
    data->g = g;
    data->ret_code = ret_code;
    *data->ret_code = I_NOTHING;
  }

  leave {
    /* do nothing */
  }

  action key_press(int key, int control, int shift) {
    if (control) {
      switch (key) {
        case 'v': goto case_paste_selection;
        case 'z':
          DO_UNDO(_owner);
          return 0;
        case 'y':
          DO_REDO(_owner);
          return 0;
        case 's':
          DO_SAVE(_owner);
          /* after a DO_XX, we must return 0 */
          return 0;
      }
      goto end_switch;
    }

    switch (key) {
      case GDK_Home:
        key = GDK_Left;
        /* fall through */
      case GDK_Left:
        /* shift + left = enter selection mode (if we can go left) */
        if (shift) {
          if (data->d->cursor.chord == 0)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_go_left;
      case GDK_End:
        key = GDK_Right;
        /* fall through */
      case GDK_Right:
        /* shift + right = enter selection mode (if we can go right) */
        if (shift) {
          if (data->d->cursor.chord == 1)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_go_right;
      case GDK_Up:
        /* shift + up = enter selection mode (if we can go up) */
        if (shift) {
          if (data->d->cursor.line == 0)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_up;
      case GDK_Down:
        /* shift + down = enter selection mode (if we can go down) */
        if (shift) {
          if (data->d->cursor.line == data->d->size - 1)
            goto end_switch;
          goto case_goto_selection;
        }
        goto case_down;
      case GDK_Page_Up: if (!shift) goto case_page_up;
      case GDK_Page_Down: if (!shift) goto case_page_down;
      case ' ': goto case_space;
      case GDK_Delete: goto case_delete;
      case GDK_BackSpace: goto case_backspace;
      case GDK_Return:
        /* shift+return to insert a text line */
        if (shift)
          goto case_insert_text;
        /* return to insert a line */
        goto case_insert_line;
      case GDK_Insert: if (shift) goto case_paste_selection; goto end_switch;
      default: goto end_switch;
    }

case_go_left:
    if (data->d->cursor.chord == 1) {
      clear_cursor(&data->d->cursor);
      data->d->cursor.chord = 0;
    }
    goto end_switch;

case_go_right:
    if (data->d->cursor.chord == 0) {
      clear_cursor(&data->d->cursor);
      data->d->cursor.chord = 1;
    }
    goto end_switch;

case_up:
    if (data->d->cursor.line == 0) goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.line--;
    data->d->cursor.chord = 0;
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;

case_down:
    if (data->d->cursor.line == data->d->size-1) goto end_switch;
    clear_cursor(&data->d->cursor);
    data->d->cursor.line++;
    data->d->cursor.chord = 0;
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;

case_page_up: {
    int height = gtk_wrapper_get_visible_height(data->d->drawing_area,
                                                data->d->zoom);
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    while (height > 0) {
      if (!data->d->cursor.line) break;
      data->d->cursor.line--;
      height -= data->d->data[data->d->cursor.line].dim.height;
    }
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;
  }

case_page_down: {
    int height = gtk_wrapper_get_visible_height(data->d->drawing_area,
                                                data->d->zoom);
    clear_cursor(&data->d->cursor);
    data->d->cursor.chord=0;
    ensure_height_in_view(data->d,
        data->d->data[data->d->cursor.line].dim.y + height);
    while (height > 0) {
      if (data->d->cursor.line == data->d->size - 1) break;
      height -= data->d->data[data->d->cursor.line].dim.height;
      data->d->cursor.line++;
    }
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;
  }

case_space:
    clear_cursor(&data->d->cursor);
    CALL(set_text_state, data->d, data->g,
         LINE_TEXT_DATA(data->d, data->d->cursor.line));
    goto end_switch;

case_delete:
    if (data->d->cursor.chord != 0) goto end_switch;
    clear_cursor(&data->d->cursor);
    history_next_is_major(&data->d->history);
    if (!delete_text_line(data->d, data->d->cursor.line)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    if (data->d->cursor.line >= data->d->size)
      data->d->cursor.line--;
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;

case_backspace:
    if (data->d->cursor.chord != 1) goto end_switch;
    clear_cursor(&data->d->cursor);
    history_next_is_major(&data->d->history);
    if (!delete_text_line(data->d, data->d->cursor.line)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    data->d->cursor.chord = 0;
    if (data->d->cursor.line >= data->d->size)
      data->d->cursor.line--;
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;

case_paste_selection:
    history_next_is_major(&data->d->history);
    clear_cursor(&data->d->cursor);
    if (!paste_selection(data->d)) {
      history_up_and_forget(data->d);
      size_post_treatment(data->d);
      CALL(error_state, data->g->win);
      goto end_switch;
    }
    size_post_treatment(data->d);
    reset_cursor(&data->d->cursor);
    set_info_chord(data->d);
    RETURN;

case_goto_selection:
    CALL(state_selection, data->d, data->g, key, data->ret_code);
    RETURN;

case_insert_text:
    /* beginning of line */
    if (data->d->cursor.chord == 0) {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_text_line(data->d, data->d->cursor.line-1)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      size_post_treatment(data->d);
    }
    /* end of line */
    else {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_text_line(data->d, data->d->cursor.line)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      data->d->cursor.chord = 0;
      size_post_treatment(data->d);
    }
    reset_cursor(&data->d->cursor);
    RETURN;

case_insert_line:
    /* beginning of line */
    if (data->d->cursor.chord == 0) {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_music_line(data->d, data->d->cursor.line-1)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      size_post_treatment(data->d);
    }
    /* end of line */
    else {
      history_next_is_major(&data->d->history);
      clear_cursor(&data->d->cursor);
      if (!insert_music_line(data->d, data->d->cursor.line)) {
        history_up_and_forget(data->d);
        size_post_treatment(data->d);
        CALL(error_state, data->g->win);
        goto end_switch;
      }
      data->d->cursor.line++;
      data->d->cursor.chord = 0;
      size_post_treatment(data->d);
    }
    reset_cursor(&data->d->cursor);
    RETURN;

end_switch:
    reset_cursor(&data->d->cursor);
  }

%%import "states/common_actions.i"
}
